/* stats.c

   Collect data for mean, max, min, &c

   Oww project
   Dr. Simon J. Melhuish

   Fri 01st December 2000
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string.h>
//#include <assert.h>

#include "wstypes.h"
#include "werr.h"
#include "stats.h"
#include "setup.h"
#include "globaldef.h"
#include "devices.h"

extern time_t the_time;

const char *stats_point_names[] = {
	"North", "NNE", "NE", "ENE",
	"East", "ESE", "SE", "SSE",
	"South", "SSW", "SW", "WSW",
	"West", "WNW", "NW", "NNW"
};

static double bearing_cos[] = {
	-0.382683432 /* Bearing is 1 - 16, so this value never used! */ ,
	0.0,
	0.382683432,
	0.707106781,
	0.923879532,
	1.0,
	0.923879532,
	0.707106781,
	0.382683432,
	0.0,
	-0.382683432,
	-0.707106781,
	-0.923879532,
	-1.0,
	-0.923879532,
	-0.707106781,
	-0.382683432
};

static double bearing_sin[] = {
	0.923879532 /* Bearing is 1 - 16, so this value never used! */ ,
	1.0,
	0.923879532,
	0.707106781,
	0.382683432,
	0.0,
	-0.382683432,
	-0.707106781,
	-0.923879532,
	-1.0,
	-0.923879532,
	-0.707106781,
	-0.382683432,
	0.0,
	0.382683432,
	0.707106781,
	0.923879532,
};

void
stats_new_anem (wsstruct vals, statsstruct * stats)
{
	/* Wind speed max/min */
	if (vals.anem_speed < stats->minWs)
		stats->minWs = vals.anem_speed;
	if (vals.anem_speed > stats->maxWs)
		stats->maxWs = vals.anem_speed;
}

void
stats_new_gpc (wsstruct vals, statsstruct * stats)
{
	int i;

	/* Maximum counter changes */
	for (i = 0; i < MAXGPC; ++i)
	{
		if (vals.gpc[i].max_sub_delta > stats->gpceventmax[i])
                  stats->gpceventmax[i] = vals.gpc[i].max_sub_delta;
	}
}

/* void stats_new_data(wsstruct vals, statsstruct *stats)

   vals - new data for integration
   stats - pointer to struct holding accrued values
*/

void
stats_new_data (wsstruct vals, statsstruct * stats)
{
	int i;

	++stats->N;
	stats->sumTime += the_time - stats->timeBase;

	/* Sum temperature readings and do max/min */
	for (i = 0; i < MAXTEMPS; ++i)
	{
		if (devices_have_thermometer (i))
		{
			++stats->T_N[i];
			stats->sumT[i] += vals.T[i];
			if (vals.T[i] < stats->minT[i])
				stats->minT[i] = vals.T[i];
			if (vals.T[i] > stats->maxT[i])
				stats->maxT[i] = vals.T[i];
		}
	}

	/* Humidity stats */
	for (i = 0; i < MAXHUMS; ++i)
	{
		if (devices_have_hum (i))
		{
			++stats->RH_N[i];
			stats->sumTrh[i] += vals.Trh[i];
			if (vals.Trh[i] < stats->minTrh[i])
				stats->minTrh[i] = vals.Trh[i];
			if (vals.Trh[i] > stats->maxTrh[i])
				stats->maxTrh[i] = vals.Trh[i];
			stats->sumRH[i] += vals.RH[i];
			if (vals.RH[i] < stats->minRH[i])
				stats->minRH[i] = vals.RH[i];
			if (vals.RH[i] > stats->maxRH[i])
				stats->maxRH[i] = vals.RH[i];
		}
	}

	/* Barometer stats */
	for (i = 0; i < MAXBAROM; ++i)
	{
		if (devices_have_barom (i))
		{
			++stats->barom_N[i];
			stats->sumTb[i] += vals.Tb[i];
			if (vals.Tb[i] < stats->minTb[i])
				stats->minTb[i] = vals.Tb[i];
			if (vals.Tb[i] > stats->maxTb[i])
				stats->maxTb[i] = vals.Tb[i];
			stats->sumbarom[i] += vals.barom[i];
			if (vals.barom[i] < stats->minbarom[i])
				stats->minbarom[i] = vals.barom[i];
			if (vals.barom[i] > stats->maxbarom[i])
				stats->maxbarom[i] = vals.barom[i];
		}
	}

	/* Sum wind x and y components, and speed */

	/* vane_bearing range is 1 to 16 i.e. 22.5 deg steps */
	/* cos and sin values are pre-calculated */

	stats->sumWx +=
		vals.anem_speed * (float) bearing_cos[vals.vane_bearing];
	stats->sumWy +=
		vals.anem_speed * (float) bearing_sin[vals.vane_bearing];
	stats->sumWs += vals.anem_speed;

	/* Record bearing too, in case of zero wind speed */
	stats->bearing = vals.vane_bearing;

	/* Wind speed max/min */
	if (vals.anem_speed < stats->minWs)
		stats->minWs = vals.anem_speed;
	if (vals.anem_speed > stats->maxWs)
		stats->maxWs = vals.anem_speed;

	/* Record monotonic rain value (tips) */
  if (devices_have_rain())
  {
    //if (stats->rainN == 0) // First rain reading this integration
	  //  stats->rainOld = stats->rainNow ;

    // rainOld collected by stats_reset()

	  stats->rainNow = vals.rain_count;
    ++stats->rainN ;
  }

	/* Record rain since last reset (inches) */
	stats->rain = vals.rain;

	for (i = 0; i < MAXGPC; ++i)
	{
		if (devices_have_gpc (i))
		{
      if (stats->gpcN[i] == 0) /* First read this integration? */
        stats->gpcOld[i] = vals.gpc[i].count_old ;

			/* Record monotonic gpc value (counts) */
			//stats->gpcOld[i] = stats->gpcNow[i] ;
			stats->gpcNow[i] = vals.gpc[i].count;
			stats->gpcEvents[i] = vals.gpc[i].event_count;

			/* Record gpc since last reset (arbitrary units) */
			stats->gpc[i] = vals.gpc[i].gpc;
      ++stats->gpcN[i] ;
		}
	}

  // Solar sensors

  for (i = 0; i < MAXSOL; ++i)
  {
    if (devices_have_solar(i))
    {
      if (stats->sol_N[i] == 0) // First value this int?
        stats->sol_min[i] = stats->sol_max[i] = vals.solar[i] ;

      ++stats->sol_N[i] ;
      stats->sol_sum[i] +=  vals.solar[i] ;
    }
  }
}

void
stats_reset (statsstruct * stats)
{
	/* Initialize the values in stats */

	unsigned long rainOld, gpcOld[MAXGPC];
	int i;
	time_t last_t = 0;

	rainOld = stats->rainNow;

	for (i = 0; i < MAXGPC; ++i)
		gpcOld[i] = stats->gpcNow[i];

	/* If this isn't the first reset, calculate last_t */
	if ((stats->N > 0) && (stats->timeBase != 0))
		last_t = stats->timeBase + stats->sumTime / stats->N;

	memset ((void *) stats, 0, sizeof (statsstruct));

	stats->rainOld = rainOld;

	stats->timeBase = the_time;
	stats->last_t = last_t;

	for (i = 0; i < MAXGPC; ++i)
  {
		stats->gpcOld[i] = gpcOld[i];
  }

	/* Set min/max outside of reasonable range */

	stats->minWs = 1e6F;
	stats->maxWs = -1e6F;

	for (i = 0; i < MAXTEMPS; ++i)
	{
		stats->minT[i] = 1e6F;
		stats->maxT[i] = -1e6F;
	}

	for (i = 0; i < MAXGPC; ++i)
	{
		stats->gpceventmax[i] = 0.0F;
	}

	for (i = 0; i < MAXHUMS; ++i)
	{
		stats->minTrh[i] = stats->minRH[i] = 1e6F;
		stats->maxTrh[i] = stats->maxRH[i] = -1e6F;
	}

	for (i = 0; i < MAXBAROM; ++i)
	{
		stats->minbarom[i] = stats->minTb[i] = 1e6F;
		stats->maxbarom[i] = stats->maxTb[i] = -1e6F;
	}
}

time_t
stats_mean_t (statsstruct * stats)
{
	if (stats->N > 0)
		return stats->timeBase + (time_t) (stats->sumTime / stats->N);
	else
		return (time_t) 0;
}

struct tm *
stats_mean_tm (statsstruct * stats)
{
	static time_t mean_time;
	mean_time = stats_mean_t (stats);
	if (mean_time == 0)
		mean_time = the_time + setup_log_interval / 2;	/* Best guess */

	return localtime (&mean_time);
}

static void
stats_closest_point (statsmean * means)
{
	/* Closest compass point */
	means->point = (int) floor (0.5 + (double) means->meanWd / 22.5);
	if (means->point < 0)
		means->point += 16;
	means->point = means->point % 16;
}

static void
stats_named_point (statsmean * means)
{
	/* Fill in the name of the compass point (in English) */

	if ((means->point < 0) || (means->point >= 16))
		means->point = 0;

	means->pointName = stats_point_names[means->point];
}

int
stats_do_means (statsstruct * stats, statsmean * means)
{
	/* Calculate the means */

	int i;

	if (stats->N < 1)
	{
		werr (0, "stats_do_means N = %d", stats->N);
		return 0;	/* Bad stats - expected N > 0 */
	}

	/* Mean time */
	means->meanTime = stats->timeBase + stats->sumTime / stats->N;
	means->meanTime_tm = *localtime (&means->meanTime);

	/* Mean temperatures */
	for (i = 0; i < MAXTEMPS; ++i)
	{
		if (stats->T_N[i] > 0)
		{
			float old_T;
			old_T = means->meanT[i];
			means->meanT[i] =
				stats->sumT[i] / (float) stats->T_N[i];
			means->haveT[i] = 1;
			stats->T_N[i] = 0;

#     ifdef HAVE_FINITEF
			if (!finitef (means->meanT[i]))
			{
				werr (WERR_WARNING,
				      "Non-finite T%d value generated in stats",
				      i + 1);
				werr (WERR_WARNING, "sumT = %f",
				      stats->sumT[i]);
				werr (WERR_WARNING, "T_N = %d",
				      stats->T_N[i]);

				means->meanT[i] = old_T;
			}
#     endif
		}
		else
		{
			means->meanT[i] = 0.0F;
			means->haveT[i] = 0;
		}
	}

	/* Mean temperatures and RH from hygrometers */
	for (i = 0; i < MAXHUMS; ++i)
	{
		if (stats->RH_N[i] > 0)
		{
			means->meanTrh[i] =
				stats->sumTrh[i] / (float) stats->RH_N[i];
			means->meanRH[i] =
				stats->sumRH[i] / (float) stats->RH_N[i];
			means->haveRH[i] = 1;
			stats->RH_N[i] = 0;
		}
		else
		{
			means->meanTrh[i] = 0.0F;
			means->meanRH[i] = 0.0F;
			means->haveRH[i] = 0;
		}
	}

	/* Mean barometer readings */
	for (i = 0; i < MAXBAROM; ++i)
	{
		if (stats->barom_N[i] > 0)
		{
			means->meanTb[i] =
				stats->sumTb[i] / (float) stats->barom_N[i];
			means->meanbarom[i] =
				stats->sumbarom[i] /
				(float) stats->barom_N[i];
			means->havebarom[i] = 1;
			stats->barom_N[i] = 0;
		}
		else
		{
			means->meanTb[i] = 0.0F;
			means->meanbarom[i] = 0.0F;
			means->havebarom[i] = 0;
		}
	}

	/* Mean wind direction */
	if (stats->sumWx != 0.0)
	{
		means->meanWd = 90.0F - 57.29578F * (float)
			atan2 ((double) stats->sumWy, (double) stats->sumWx);
	}
	else
	{
		/* Zero wind speed - just use last value of vane */
		means->meanWd = (float) (stats->bearing - 1) * 22.5F;
	}

	/* Wind gust */
	means->maxWs = stats->maxWs;

	/* Confine to range 0 to 360 degrees */
	if (means->meanWd < 0)
		means->meanWd += 360.0F;

	/* Closest compass point */
	stats_closest_point (means);

	/* Named compass point */
	stats_named_point (means);

	/* Mean wind speed */
	means->meanWs = stats->sumWs / (float) stats->N;

	/* Rainfall since last reset */
	means->rain = stats->rain;

	/* Rainfall this integration */
	if ((stats->rainN > 0) && (stats->rainNow > stats->rainOld))
	{
		means->deltaRain =
			RAINCALIB * (float) (stats->rainNow - stats->rainOld);
		if (stats->last_t > 0)
			means->rain_rate = 3600.0F * means->deltaRain /
				(float) (the_time - stats->last_t);
		else
			means->rain_rate = 0.0F;
	}
	else
	{
		means->deltaRain = means->rain_rate = 0.0F;
	}

	/* GPCs */
	for (i = 0; i < MAXGPC; ++i)
	{
		if (stats->gpcN[i] > 0)
		{
      int delta ;

			means->haveGPC[i] = 1;
			means->gpc[i] = stats->gpc[i];
      means->gpcmono[i] = stats->gpcNow[i] ;
			means->gpcEvents[i] = stats->gpcEvents[i] ;
                        means->gpceventmax[i] = stats->gpceventmax[i] ;

			if (stats->gpcNow[i] > stats->gpcOld[i])
      {
        delta = stats->gpcNow[i] - stats->gpcOld[i] ;
	means->deltagpc[i] =
	  devices_list[devices_GPC1 + i].calib[0] * (float) delta ;
        means->gpcrate[i] = 3600.0F * means->deltagpc[i] /
				  (float) (the_time - stats->last_t);
      }
			else
      {
				means->deltagpc[i] = means->gpcrate[i] = 0.0F;
      }
		}
		else
		{
			means->haveGPC[i] = 0;
      means->gpc[i] = 0 ;
      means->gpcmono[i] = 0 ;
      means->deltagpc[i] = means->gpcrate[i] = 0.0F;
		}
	}

  // Solar sensors

  for (i = 0; i < MAXSOL; ++i)
  {
    if (stats->sol_N[i] > 0)
    {
      means->meansol[i] = stats->sol_sum[i] / (float) stats->sol_N[i] ;
      means->havesol[i] = 1 ;
    }
    else
    {
      means->meansol[i] = 0.0F ;
      means->havesol[i] = 0 ;
    }
  }

  return 1;		/* OK */
}

int
stats_do_ws_means (wsstruct * vals, statsmean * means)
{
	/* Use ws vals in place of means the means */

	int i;

	/* Mean time */
	means->meanTime = the_time;
	means->meanTime_tm = *localtime (&means->meanTime);

	/* Mean temperatures */
	for (i = 0; i < MAXTEMPS; ++i)
	{
		if (devices_have_thermometer (i))
		{
			means->meanT[i] = vals->T[i];
			means->haveT[i] = 1;
		}
		else
		{
			means->meanT[i] = 0.0F;
			means->haveT[i] = 0;
		}
	}

	/* Mean humidities */
	for (i = 0; i < MAXHUMS; ++i)
	{
		if (devices_have_hum (i))
		{
			means->meanTrh[i] = vals->Trh[i];
			means->meanRH[i] = vals->RH[i];
			means->haveRH[i] = 1;
		}
		else
		{
			means->meanTrh[i] = means->meanRH[i] = 0.0F;
			means->haveRH[i] = 0;
		}
	}

	/* Mean barometer readings */
	for (i = 0; i < MAXBAROM; ++i)
	{
		if (devices_have_barom (i))
		{
			means->meanTb[i] = vals->Tb[i];
			means->meanbarom[i] = vals->barom[i];
			means->havebarom[i] = 1;
		}
		else
		{
			means->meanbarom[i] = means->meanTb[i] = 0.0F;
			means->havebarom[i] = 0;
		}
	}

	/* Wind bearing and point */
	means->point = vals->vane_bearing - 1;
	if ((means->point < 0) || (means->point > 15))
	{
		werr (WERR_DEBUG0, "stats_do_ws_means() Wind point = %d",
		      means->point);
		means->point = 0;
	}
	means->meanWd = (float) means->point * 22.5F;

	/* Named compass point */
	stats_named_point (means);

	/* Mean wind speed */
	means->meanWs = vals->anem_speed;

	/* Wind gust - biggest value found during update interval */
	means->maxWs = vals->anem_int_gust;

	/* Rainfall since last reset */
  if (devices_have_rain())
  {
    means->rain = vals->rain;

    /* Rainfall rate */
    means->rain_rate = vals->rain_rate;

    /* Rainfall this integration */
    if (vals->rain_count > vals->rain_count_old)
      means->deltaRain = RAINCALIB *
        (float) (vals->rain_count - vals->rain_count_old);
    else
      means->deltaRain = 0.0F;
  }
  else
  {
    means->rain = means->rain_rate = means->deltaRain = 0.0F ;
  }

	for (i = 0; i < MAXGPC; ++i)
	{
		if (devices_have_gpc (i))
		{
			/* GPC since last reset */
			means->gpc[i] = vals->gpc[i].gpc;
			means->haveGPC[i] = 1;
      means->gpcmono[i] = vals->gpc[i].count ;
      means->gpcrate[i] = vals->gpc[i].rate ;
			means->gpcEvents[i] = vals->gpc[i].event_count;

			/* GPC increment this integration */
			if (vals->gpc[i].count > vals->gpc[i].count_old)
                        {
				means->deltagpc[i] =
					devices_list[devices_GPC1 +
						     i].calib[0] *
					(float) (vals->gpc[i].count -
						 vals->gpc[i].count_old);
                                means->gpceventmax[i] = vals->gpc[i].max_sub_delta ;
                        }
			else
                          means->deltagpc[i] = means->gpceventmax[i] = 0.0F;
		}
		else
		{
			means->gpc[i] = means->deltagpc[i] = 0.0F;
			means->haveGPC[i] = 0;
		}
	}

  // Solar sensors

  for (i = 0; i < MAXSOL; ++i)
  {
    if (devices_have_solar(i))
    {
      means->meansol[i] = vals->solar[i] ;
      means->havesol[i] = 1 ;
    }
    else
    {
       means->meansol[i] = 0.0F ;
       means->havesol[i] = 0 ;
    }
  }

  return 1;		/* OK */
}

